@@ -1,17 +1,23 @@
 import functools
 from collections import defaultdict
+from logging import getLogger
 from pathlib import Path
 from typing import *
 
+from orkestra.exceptions import CompositionError
 from orkestra.interfaces import (
     Duration,
     Runtime,
     LambdaInvocationType,
     IntegrationPattern,
     Tracing,
+    StateMachineType as SfnType,
+    PythonLayerVersion,
 )
-from orkestra.utils import coerce, cdk_patch
-from orkestra.exceptions import CompositionError
+from orkestra.utils import coerce, _coalesce
+
+logger = getLogger(__name__)
+
 
 OptionalFn = Optional[Union[Callable, Iterable[Callable]]]
 
@@ -31,22 +37,44 @@ def _incremental_id(id):
 
 
 class Compose:
+    _powertools_defaults = {
+        "log_event": True,
+        "capture_response": True,
+        "capture_error": True,
+        "raise_on_empty_metrics": False,
+        "capture_cold_start_metric": True,
+    }
+
+    _sfn_task_defaults = {
+        "payload_response_only": True,
+    }
+
+    @property
+    def _aws_lambda_defaults(self):
+        from aws_cdk.aws_lambda import Runtime, Tracing
+
+        return {
+            "runtime": Runtime.PYTHON_3_8,
+            "tracing": Tracing.ACTIVE,
+        }
+
     def __init__(
         self,
         func: OptionalFn = None,
         enable_powertools: bool = False,
-        log_event=True,
-        capture_response=True,
-        capture_error=True,
-        raise_on_empty_metrics=False,
-        capture_cold_start_metric=True,
-        default_dimensions=None,
-        model=None,
-        envelope=None,
-        timeout: Optional[Duration] = None,
         is_map_job: bool = False,
         capture_map_errors: bool = False,
+        log_event: Optional[bool] = None,
+        capture_response: Optional[bool] = None,
+        capture_error: Optional[bool] = None,
+        raise_on_empty_metrics: Optional[bool] = None,
+        capture_cold_start_metric: Optional[bool] = None,
+        default_dimensions: Optional[dict] = None,
+        model: Optional["pydantic.BaseModel"] = None,
+        envelope: Optional["pydantic.BaseModel"] = None,
+        timeout: Optional[Duration] = None,
         runtime: Optional[Runtime] = None,
+        layers: Optional[Sequence[PythonLayerVersion]] = None,
         comment: Optional[str] = None,
         input_path: Optional[str] = None,
         items_path: Optional[str] = None,
@@ -54,7 +82,7 @@ class Compose:
         output_path: Optional[str] = None,
         parameters: Optional[Mapping[str, Any]] = None,
         result_path: Optional[str] = None,
-        result_selector: Optional[Mapping[str, any]] = None,
+        result_selector: Optional[Mapping[str, Any]] = None,
         client_context: Optional[str] = None,
         invocation_type: Optional[LambdaInvocationType] = None,
         payload_response_only: Optional[bool] = None,
@@ -64,6 +92,8 @@ class Compose:
         integration_pattern: Optional[IntegrationPattern] = None,
         sfn_timeout: Optional[Duration] = None,
         tracing: Optional[Tracing] = None,
+        state_machine_type: Optional[SfnType] = None,
+        state_machine_name: Optional[str] = None,
         **aws_lambda_constructor_kwargs,
     ):
         """
@@ -82,6 +112,7 @@ class Compose:
             model: passed to aws_lambda_powertools.utilities.parser.event_parser
             envelope: passed to aws_lambda_powertools.utilities.parser.event_parser
             runtime: the python runtime to use for the lambda
+            layers: A list of layers to add to the function’s execution environment. You can configure your Lambda function to pull in additional code during initialization in the form of layers. Layers are packages of libraries or other dependencies that can be used by multiple functions. Default: - No layers.
             is_map_job: whether the lambda is a map job
             capture_map_errors: set true to add guarantee successful map job execution
             comment: An optional description for this state. Default: No comment
@@ -101,6 +132,8 @@ class Compose:
             integration_pattern: AWS Step Functions integrates with services directly in the Amazon States Language. You can control these AWS services using service integration patterns Default: IntegrationPattern.REQUEST_RESPONSE
             sfn_timeout: Timeout for the state machine. Default: - None
             tracing: Enable AWS X-Ray Tracing for Lambda Function. Default: Tracing.Enabled
+            state_machine_type: Type of the state machine. Default: StateMachineType.STANDARD
+            state_machine_name: A name for the state machine. Default: A name is automatically generated
             **aws_lambda_constructor_kwargs: pass directly to sfn.PythonFunction
 
         For cdk params see https://docs.aws.amazon.com/cdk/api/latest/python/modules.html
@@ -126,44 +159,56 @@ class Compose:
             "parameters": parameters,
         }
 
-        self.lambda_invoke_kwargs = {
-            "client_context": client_context,
-            "invocation_type": invocation_type,
-            "payload_response_only": payload_response_only,
-            "retry_on_service_exceptions": retry_on_service_exceptions,
-            "heartbeat": heartbeat,
-            "integration_pattern": integration_pattern,
-            "timeout": sfn_timeout,
-            "comment": comment,
-            "input_path": input_path,
-            "output_path": output_path,
-            "result_path": result_path,
-            "result_selector": result_selector,
-            "qualifier": qualifier,
-        }
+        self.lambda_invoke_kwargs = _coalesce(
+            self._sfn_task_defaults,
+            {
+                "client_context": client_context,
+                "invocation_type": invocation_type,
+                "payload_response_only": payload_response_only,
+                "retry_on_service_exceptions": retry_on_service_exceptions,
+                "heartbeat": heartbeat,
+                "integration_pattern": integration_pattern,
+                "timeout": sfn_timeout,
+                "comment": comment,
+                "input_path": input_path,
+                "output_path": output_path,
+                "result_path": result_path,
+                "result_selector": result_selector,
+                "qualifier": qualifier,
+            },
+        )
 
-        self.powertools_kwargs = {
-            "log_event": log_event,
-            "capture_error": capture_error,
-            "capture_response": capture_response,
-            "capture_cold_start_metric": capture_cold_start_metric,
-            "raise_on_empty_metrics": raise_on_empty_metrics,
-            "default_dimensions": default_dimensions,
-            "model": model,
-            "envelope": envelope,
+        self.powertools_kwargs = _coalesce(
+            self._powertools_defaults,
+            default_dimensions=default_dimensions,
+            model=model,
+            envelope=envelope,
+            log_event=log_event,
+            capture_response=capture_response,
+            capture_error=capture_error,
+            raise_on_empty_metrics=raise_on_empty_metrics,
+            capture_cold_start_metric=capture_cold_start_metric,
+        )
+
+        self.state_machine_kwargs = {
+            "state_machine_type": state_machine_type,
+            "state_machine_name": state_machine_name,
         }
 
         self.aws_lambda_constructor_kwargs.update(
             timeout=timeout,
             runtime=runtime,
             tracing=tracing,
+            layers=layers,
         )
 
         self.enable_powertools = enable_powertools
 
-        self.update_metadata()
+        self._lambda_function = None
 
-    def update_metadata(self):
+        self._update_metadata()
+
+    def _update_metadata(self):
         if self.func and not isinstance(self.func, (list, tuple)):
             module = self.func.__module__.split(".")
             self.aws_lambda_constructor_kwargs.update(
@@ -174,9 +219,9 @@ class Compose:
 
     def __call__(self, event, context=None):
 
-        if isinstance(self.func, (list, tuple)):
+        if self.func is not None and not callable(self.func):
 
-            raise TypeError("can't call a list of functions")
+            raise TypeError(f"{self.func} is not callable")
 
         if self.func is not None:
 
@@ -195,16 +240,16 @@ class Compose:
 
             self.func = event
 
-            self.update_metadata()
+            self._update_metadata()
 
             return self
 
     def __repr__(self) -> str:
 
-        if isinstance(self.func, (list, tuple)):
-            func = repr(self.func)
+        if hasattr(self.func, "__name__"):
+            func = self.func.__name__
         else:
-            func = self.func.__name__ if self.func is not None else None
+            func = repr(self.func)
 
         return f"Compose(func={func})"
 
@@ -223,50 +268,61 @@ class Compose:
         **kwargs,
     ):
 
-        from aws_cdk import aws_lambda, aws_lambda_python
-
-        keyword_args = {
-            **composable.aws_lambda_constructor_kwargs,
-            **kwargs,
-        }
-
-        cdk_patch(keyword_args)
-
-        if keyword_args.get("runtime") is None:
-
-            keyword_args.update(
-                runtime=aws_lambda.Runtime.PYTHON_3_8,
-            )
-
-        if keyword_args.get("tracing") is None:
-
-            keyword_args.update(
-                tracing=aws_lambda.Tracing.ACTIVE,
-            )
+        from aws_cdk.aws_lambda_python import PythonFunction
 
         id = id or _incremental_id(composable.func.__name__ + "_fn")
 
-        keyword_args = {k: v for k, v in keyword_args.items() if v is not None}
+        layers = composable.aws_lambda_constructor_kwargs.get("layers")
 
-        return aws_lambda_python.PythonFunction(
+        if layers is not None:
+
+            runtime = _coalesce(
+                composable._aws_lambda_defaults,
+                composable.aws_lambda_constructor_kwargs,
+            )["runtime"]
+
+            layers = [
+                l.cdk_construct(
+                    scope,
+                    _incremental_id(f"{id}_layer"),
+                    compatible_runtimes=[runtime],
+                )
+                for l in layers
+            ]
+
+            if kwargs.get("layers") is not None:
+
+                kwargs["layers"].extend(layers)
+
+            else:
+
+                kwargs["layers"] = layers
+
+        kwargs = _coalesce(
+            composable._aws_lambda_defaults,
+            composable.aws_lambda_constructor_kwargs,
+            kwargs,
+        )
+
+        return PythonFunction(
             scope,
             id,
-            **keyword_args,
+            **kwargs,
         )
 
     def aws_lambda(
         self,
-        scope,
-        id=None,
+        scope: "aws_cdk.core.Construct",
+        id: Optional[str] = None,
         **kwargs,
     ):
         """
         Return lambda cdk construct.
 
         Args:
-            scope:
-            id:
-            **kwargs:
+            scope: cdk construct
+            id: construct id
+            **kwargs: to be passed to aws_cdk.aws_lambda_python.PythonFunction
 
         Returns (aws_cdk.aws_lambda_python.PythonFunction): python lambda
 
@@ -281,10 +337,10 @@ class Compose:
 
     def task(
         self,
-        scope,
-        id=None,
-        payload_response_only=True,
-        function_name=None,
+        scope: "aws_cdk.core.Construct",
+        id: Optional[str] = None,
+        payload_response_only: bool = True,
+        function_name: Optional[str] = None,
         **kwargs,
     ):
         """
@@ -298,34 +354,22 @@ class Compose:
 
             id = id or _incremental_id(self.func.__name__)
 
-            map_kwargs = dict(id=id)
+            map_kwargs = _coalesce(self.map_job_kwargs)
 
-            map_kwargs.update(
-                {k: v for k, v in self.map_job_kwargs.items() if v is not None}
-            )
-
-            task = sfn.Map(scope, **map_kwargs)
+            task = sfn.Map(scope, id, **map_kwargs)
 
             self.lambda_fn = self.aws_lambda(
                 scope,
                 function_name=function_name,
             )
 
-            keyword_args = dict(
+            keyword_args = _coalesce(
+                self.lambda_invoke_kwargs,
                 lambda_function=self.lambda_fn,
                 payload_response_only=payload_response_only,
+                **kwargs,
             )
 
-            keyword_args.update(
-                {
-                    k: v
-                    for k, v in self.lambda_invoke_kwargs.items()
-                    if v is not None
-                }
-            )
-
-            cdk_patch(keyword_args)
-
             task_id = f"invoke_{id}"
 
             invoke_lambda = sfn_tasks.LambdaInvoke(
@@ -344,39 +388,7 @@ class Compose:
 
             task.iterator(invoke_lambda)
 
-        elif not isinstance(self.func, (list, tuple)):
-
-            id = id or _incremental_id(self.func.__name__)
-
-            self.lambda_fn = self.aws_lambda(
-                scope,
-                function_name=function_name,
-            )
-
-            keyword_args = dict(
-                lambda_function=self.lambda_fn,
-                payload_response_only=payload_response_only,
-            )
-
-            keyword_args.update(kwargs)
-
-            keyword_args.update(
-                {
-                    k: v
-                    for k, v in self.lambda_invoke_kwargs.items()
-                    if v is not None
-                }
-            )
-
-            cdk_patch(keyword_args)
-
-            task = sfn_tasks.LambdaInvoke(
-                scope,
-                id,
-                **keyword_args,
-            )
-
-        else:
+        elif isinstance(self.func, (list, tuple)):
 
             id = "parallelize " + (
                 "".join([c.func.__name__ for c in self.func])
@@ -391,23 +403,13 @@ class Compose:
 
                 lambda_fn = fn.aws_lambda(scope)
 
-                keyword_args = dict(
+                keyword_args = _coalesce(
+                    self.lambda_invoke_kwargs,
                     lambda_function=lambda_fn,
                     payload_response_only=payload_response_only,
+                    **kwargs,
                 )
 
-                keyword_args.update(kwargs)
-
-                keyword_args.update(
-                    {
-                        k: v
-                        for k, v in self.lambda_invoke_kwargs.items()
-                        if v is not None
-                    }
-                )
-
-                cdk_patch(keyword_args)
-
                 branch = sfn_tasks.LambdaInvoke(
                     scope,
                     _incremental_id(fn.func.__name__),
@@ -425,20 +427,52 @@ class Compose:
 
                 task.branch(branch)
 
+        else:
+
+            id = id or _incremental_id(self.func.__name__)
+
+            self.lambda_fn = self.aws_lambda(
+                scope,
+                function_name=function_name,
+            )
+
+            keyword_args = _coalesce(
+                self.lambda_invoke_kwargs,
+                lambda_function=self.lambda_fn,
+                payload_response_only=payload_response_only,
+                **kwargs,
+            )
+
+            task = sfn_tasks.LambdaInvoke(
+                scope,
+                id,
+                **keyword_args,
+            )
+
         return coerce(task)
 
     def definition(
         self,
-        scope,
-        definition=None,
-        previous=None,
+        scope: "aws_cdk.core.Construct",
+        previous_definition: Optional[
+            "aws_cdk.aws_stepfunctions.IChainable"
+        ] = None,
+        previously_composed: Optional["Compose"] = None,
     ):
         """
         Return automagically composed cdk state machine definition.
-        """
-        previous = previous or []
 
-        if self in previous:
+        Args:
+            scope: cdk scope
+            previous_definition: the previous definition
+            previously_composed: the previously composed
+
+        Returns:
+
+        """
+        previously_composed = previously_composed or []
+
+        if self in previously_composed:
 
             raise CompositionError(
                 f"Failed to compose {self}. Composition using >> must be acyclic."
@@ -446,7 +480,11 @@ class Compose:
 
         task = self.task(scope)
 
-        definition = task if definition is None else definition.next(task)
+        definition = (
+            task
+            if previous_definition is None
+            else previous_definition.next(task)
+        )
 
         if self.downstream:
 
@@ -454,51 +492,61 @@ class Compose:
 
                 c.definition(
                     scope,
-                    definition=definition,
-                    previous=previous + [self],
+                    previous_definition=definition,
+                    previously_composed=previously_composed + [self],
                 )
 
         return definition
 
     def state_machine(
         self,
-        scope,
-        id=None,
-        tracing_enabled=True,
-        state_machine_name=None,
+        scope: "aws_cdk.core.Construct",
+        id: Optional[str] = None,
+        tracing_enabled: bool = True,
+        state_machine_name: Optional[str] = None,
+        state_machine_type: Optional[
+            Union[SfnType, "aws_cdk.aws_stepfunctions.StateMachineType"]
+        ] = None,
         **kwargs,
     ):
         """
         Return step functions state machine cdk construct.
 
         Args:
-            scope:
-            id:
-            tracing_enabled:
-            state_machine_name:
+            scope: cdk scope
+            id: cdk id
+            tracing_enabled: xray tracing
+            state_machine_name: name of state machine
+            state_machine_type: express or standard
             **kwargs:
 
         Returns:
 
         """
 
-        from aws_cdk import aws_stepfunctions as sfn
+        from aws_cdk.aws_stepfunctions import StateMachine
 
         id = id or _incremental_id(f"{self.func.__name__}_sfn")
 
-        return sfn.StateMachine(
+        state_machine_kwargs = _coalesce(
+            self.state_machine_kwargs,
+            state_machine_name=state_machine_name,
+            state_machine_type=state_machine_type,
+            **kwargs,
+        )
+
+        return StateMachine(
             scope,
             id,
             definition=self.definition(scope),
             tracing_enabled=tracing_enabled,
-            state_machine_name=state_machine_name,
-            **kwargs,
+            **state_machine_kwargs,
         )
 
     def schedule(
         self,
-        scope,
-        id=None,
+        scope: "aws_cdk.core.Construct",
+        id: Optional[str] = None,
         expression: Optional[str] = None,
         day: Optional[str] = None,
         hour: Optional[str] = None,
@@ -506,30 +554,28 @@ class Compose:
         month: Optional[str] = None,
         week_day: Optional[str] = None,
         year: Optional[str] = None,
-        function_name=None,
-        state_machine_name=None,
-        dead_letter_queue_enabled=False,
+        state_machine_name: Optional[str] = None,
+        state_machine_type: Optional[SfnType] = None,
         **kwargs,
-    ):
+    ) -> tuple:
         """
         Schedule lambda or state machine to run on interval using EventBridge scheduled event rule.
 
         Args:
-            scope:
-            id:
-            expression:
-            day:
-            hour:
-            minute:
-            month:
-            week_day:
-            year:
-            function_name:
-            state_machine_name:
-            dead_letter_queue_enabled:
+            scope: cdk scope
+            id: construct id
+            expression: interval at which to run. Can be cron expression or CloudWatch rate expression
+            day: day of month
+            hour: hour of day
+            minute: minute of our
+            month: month of year
+            week_day: week day
+            year: year
+            state_machine_name: the state machine name, if downstream
+            state_machine_type: type of state machine; express or standard
             **kwargs:
 
-        Returns:
+        Returns (tuple): EventBridge schedule rule, SFN State Machine
 
         """
         from aws_cdk import aws_events as eventbridge
@@ -556,35 +602,30 @@ class Compose:
             **kwargs,
         )
 
-        if not self.downstream:
-            fn = self.aws_lambda(
-                scope,
-                function_name=function_name,
-                dead_letter_queue_enabled=dead_letter_queue_enabled,
-            )
-            target = eventbridge_targets.LambdaFunction(handler=fn)
-        else:
-            state_machine = self.state_machine(
-                scope,
-                state_machine_name=state_machine_name,
-            )
-            target = eventbridge_targets.SfnStateMachine(machine=state_machine)
+        state_machine = self.state_machine(
+            scope,
+            state_machine_name=state_machine_name,
+            state_machine_type=state_machine_type,
+            **kwargs,
+        )
+
+        target = eventbridge_targets.SfnStateMachine(machine=state_machine)
 
         rule.add_target(target)
 
-        return rule
+        return rule, state_machine
 
 
 def powertools(
-    decorated=None,
-    log_event=True,
-    capture_response=True,
-    capture_error=True,
-    raise_on_empty_metrics=False,
-    capture_cold_start_metric=True,
-    default_dimensions=None,
-    model=None,
-    envelope=None,
+    decorated: Optional[Callable] = None,
+    log_event: bool = True,
+    capture_response: bool = True,
+    capture_error: bool = True,
+    raise_on_empty_metrics: bool = False,
+    capture_cold_start_metric: bool = True,
+    default_dimensions: Optional[dict] = None,
+    model: Optional["pydantic.BaseModel"] = None,
+    envelope: Optional["pydantic.BaseModel"] = None,
 ):
     """
     AWS lambda powertools shortcut.
@@ -655,6 +696,8 @@ def powertools(
                 Exists because event_parser expects the function it wraps to be named "handler".
                 """
 
+                logger.debug(f"parsing {event = }")
+
                 parsed_event = parse(
                     event=event,
                     model=model,
